import logging
from sqlalchemy import create_engine  # and_
from sqlalchemy.orm import sessionmaker
from sqlalchemy.engine import reflection
# from sqlalchemy.exc import SQLAlchemyError, IntegrityError
from .models import Base, Statement, Tag, TagAssociationStatement, AccessLog
from .exceptions import DeleteDataWithoutConditionError, ModelNotExistError, ExecuteSqlError
from .constants import DEFAULT_DATABASE_URI


class SQLStorage:
    """
    The SQLStorageAdapter allows ChatterBot to store conversation
    data in any database supported by the SQL Alchemy ORM.

    All parameters are optional, by default a sqlite database is used.

    It will check if tables are present, if they are not, it will attempt
    to create the required tables.

    :keyword database_uri: eg: sqlite:///database_test.sqlite3',
        The database_uri can be specified to choose database driver.
    :type database_uri: str
    """

    def __init__(self, **kwargs):
        self.logger = kwargs.get('logger', logging.getLogger(__name__))
        self.database_uri = kwargs.get('database_uri', None)

        # Create a file database if the database is not a connection string
        if not self.database_uri:
            self.database_uri = DEFAULT_DATABASE_URI

        # connect db
        self.engine = create_engine(self.database_uri, encoding='utf8', echo=False)
        self.Session = sessionmaker(bind=self.engine, expire_on_commit=True)
        if not self.engine.dialect.has_table(self.engine, 'statement'):
            self.create_database()

    def get_model(self, model_name):
        """
        Return the model class for a given model name.

        model_name is case insensitive.
        """
        get_model_method = getattr(self, 'get_{}_model'.format(model_name.lower()), None)

        if not get_model_method:
            raise ModelNotExistError(model_name)

        return get_model_method()

    @staticmethod
    def get_statement_model():
        """
        Return the statement model class
        """
        return Statement

    @staticmethod
    def get_tag_model():
        """
        Return the tag model class
        """
        return Tag

    @staticmethod
    def get_tag_association_statement_model():
        """
        Return the tag_association_statement model class
        """
        return TagAssociationStatement

    @staticmethod
    def get_access_log_model():
        """
        Return the tag_association_statement model class
        """
        return AccessLog

    def create_database(self):
        """
        Populate the database with the tables.
        """
        Base.metadata.create_all(self.engine)

    def count(self, model_name):
        """
        Return the number of entries in the database.
        """
        model = self.get_model(model_name)

        session = self.Session()
        model_count = session.query(model).count()
        session.close()
        return model_count

    def create(self, model_name, **kwargs):
        """
        add data to the database
        """
        model = self.get_model(model_name)
        session = self.Session()
        model_new_data = model(**kwargs)
        session.add(model_new_data)
        session.commit()
        id_ = model_new_data.id
        session.close()
        return id_
        # return list(self.filter(model_name, id=id))[0]

    def delete(self, model_name, **kwargs):
        """
        delete matching data from the database
        """
        # Report error without query conditions
        if not kwargs:
            raise DeleteDataWithoutConditionError()
        model = self.get_model(model_name)
        session = self.Session()
        session.query(model).filter_by(**kwargs).delete()
        session.commit()
        session.close()

    def filter(self, model_name, to_dict=False, **kwargs):
        model = self.get_model(model_name)
        session = self.Session()
        all_filter_object = session.query(model).filter_by(**kwargs)
        if to_dict:
            inspect = reflection.Inspector.from_engine(self.engine)
            all_colum_name = [colum_info['name'] for colum_info in inspect.get_columns(model_name)]
            for object_ in all_filter_object:
                object_data = {}
                for colum_name in all_colum_name:
                    object_data[colum_name] = getattr(object_, colum_name)
                yield object_data
        else:
            for object_ in all_filter_object:
                yield object_

        session.close()

    def all(self, model_name):
        return self.filter(model_name)

    def execute(self, sql):
        """execute sql statement

        :param str sql: sql statement

        return(tuple): the first element of the tuple is the number of rows affected by the sql statement,
                       if it is a select statement, the second element returns the data,if it is not a  select
                       statement,the second element return null
        """
        session = self.Session()

        try:
            execute_result = session.execute(sql)
        except Exception as e:
            raise ExecuteSqlError('execute sql "%s" failed, %s' % (sql, e))

        # return the data if it is a select statement
        if execute_result.returns_rows:
            ret = execute_result.fetchall()
        else:
            ret = execute_result.rowcount

        session.commit()
        session.close()
        return ret

    # @staticmethod
    # def _session_commit(session):
    #     try:
    #         session.commit()
    #     except SQLAlchemyError as e:
    #         raise StorageError(str(e))


class StatementStorage:
    def __init__(self, **kwargs):
        database_uri = kwargs.get('database_uri', None)
        self.db = SQLStorage(database_uri=database_uri)
        self._operating_model = kwargs.get('operating_model', 'statement')

    @property
    def operating_model(self):
        return self._operating_model

    @operating_model.setter
    def operating_model(self, new_operating_model):
        self._operating_model = new_operating_model

    def count(self):
        return self.db.count(self.operating_model)

    def delete(self, **kwargs):
        return self.db.delete(self.operating_model, **kwargs)

    def create(self, **kwargs):
        return self.db.create(self.operating_model, **kwargs)

    def filter(self, **kwargs):
        return self.db.filter(self.operating_model, **kwargs)

    def all(self):
        return self.db.filter(self.operating_model)



import paramiko
import chardet


class Connection:
    def __init__(self, host, user, pwd, **kwargs):
        self.ssh_client = paramiko.SSHClient()
        self.ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        port = kwargs.get('port', 22)
        timeout = kwargs.get('timeout', 3)
        self.ssh_client.connect(host, port, user, pwd, timeout=timeout)

    def execute(self, cmd):
        # stdin, self.stdout, self.stderr = self.client.exec_command(cmd)
        channel = self.ssh_client._transport.open_session()
        channel.exec_command(cmd)
        output = channel.makefile('rb', -1).read()
        guess_coding = chardet.detect(output)['encoding'] or 'ascii'
        output = output.decode(guess_coding)
        return channel.exit_status, output

    def __del__(self):
        self.close()

    def close(self):
        try:
            self.ssh_client.close()
        except:
            pass


def get_disk_usage(ip, username, password):
    """
    :param str ip: 主机的ip地址
    :param str username: 登录的用户名
    :param str password: 用户名所使用的密码
    :return:
    """
    ssh = Connection(ip, username, password)
    disk_usage = ssh.execute('df -h')[1]
    print(disk_usage)
    ssh.close()
    return disk_usage


def get_memory_usage(ip, username, password):
    """
    :param str ip: 主机的ip地址
    :param str username: 登录的用户名
    :param str password: 用户名所使用的密码
    :return:
    """
    ssh = Connection(ip, username, password)
    return ssh.execute('free')[1]


# if __name__ == '__main__':
#     print(get_disk_usage('10.1.196.29', 'root', '123456'))




# from datetime import datetime
# from dateutil import parser as date_parser
# from .extractor import Extractor
from .utils import import_module, get_function_arguments, logger, get_function_parameter_desc, get_object_path
from .exceptions import NotEnoughParameterError


class StatementMixin:
    """
    This class has shared methods used to
    normalize different statement models.
    """

    statement_field_names = [
        'id',
        'question',
        'reference_question',
        'confidence',
        'answer',
        'category',
        'type',
        'parameters',
        'extractor',
    ]

    def get_statement_field_names(self):
        """
        Return the list of field names for the statement.
        """
        return self.statement_field_names

    def serialize(self):
        """
        :returns: A dictionary representation of the statement object.
        :rtype: dict
        """
        data = {}

        for field_name in self.get_statement_field_names():
            # field_method = getattr(self, 'get_{}'.format(
            #     field_name
            # ), None)
            #
            # if field_method:
            #     data[field_name] = field_method()
            # else:
            data[field_name] = getattr(self, field_name)

        return data


class Statement(StatementMixin):
    """
    A statement represents a single spoken entity, sentence or
    phrase that someone can say.
    """
    def __init__(self, question, answer=None, **kwargs):
        self.question = str(question)
        self.answer = answer
        self.id = kwargs.get('id', -1)
        self.reference_question = kwargs.get('reference_question', '')
        self.category = kwargs.get('category', '其他')

        # statement type, if the type is 0, answer = answer, if it is 1, then answer = answer(**parameters)
        self.type = kwargs.get('type', 0)

        # get parameter
        parameters = kwargs.get('parameters', {})
        if self.type == 1:
            self.answer_fun = import_module(self.answer)
            self.parameters = get_function_arguments(self.answer_fun)
            if isinstance(parameters, str):
                for parameter in parameters.split(';'):
                    if parameter.strip():
                        name, value = parameter.strip().split('=')
                        if name.strip() in self.parameters:
                            self.parameters[name.strip()] = value.strip()
            elif hasattr(parameters, 'items'):
                for name, value in getattr(parameters, 'items')():
                    if name in self.parameters:
                        self.parameters[name] = value
        else:
            self.parameters = parameters

        # get extractor
        self.extractor = kwargs.get('extractor', 'chatbot.extractor.Extractor') or 'chatbot.extractor.Extractor'

        # self.created_at = kwargs.get('created_at', datetime.now())
        # if not isinstance(self.created_at, datetime):
        #     self.created_at = date_parser.parse(self.created_at)

        # This is the confidence with which the chat bot believes
        # this is an accurate answer. This value is set when the
        # statement is returned by the chat bot.
        self.confidence = kwargs.get('confidence', 0)

        # logger
        self.logger = kwargs.get('logger', logger)

    def __str__(self):
        return self.question

    def __repr__(self):
        return '<Statement text:%s>' % self.question

    def get_answer(self):
        if self.type == 0:
            return self.answer
        elif self.type == 1:
            for parameter, value in self.parameters.items():
                if value is None:
                    parameter_desc = get_function_parameter_desc(self.answer_fun).get(parameter, parameter)
                    raise NotEnoughParameterError(parameter, parameter_desc)
            try:
                self.logger.info('Call the function "{}" to get the answer, all the parameters of the '
                                 'function are "{}"'.format(get_object_path(self.answer_fun), self.parameters))
                text = self.answer_fun(**self.parameters)
            except Exception as e:
                text = str(e)
            return text



import importlib
import inspect
import logging
import sys
import os
import re
import collections
import jieba.analyse
from .exceptions import InvalidTypeError
from .constants import PROJECT_DIR_PATH, LOG_FILE_NAME, LOG_LEVEL, APP_NAME


def import_module(dotted_path='re.search'):
    """
    Imports the specified module based on the
    dot notated import path for the module.
    """
    module_parts = dotted_path.split('.')
    module_path = '.'.join(module_parts[:-1])
    module = importlib.import_module(module_path)

    return getattr(module, module_parts[-1])


def initialize_class(data, *args, **kwargs):
    """
    :param data: A string or dictionary containing a import_path attribute.
    """
    if isinstance(data, dict):
        import_path = data.pop('import_path')
        kwargs.update(data)
        class_ = import_module(import_path)
    else:
        class_ = import_module(data)

    return class_(*args, **kwargs)


def validate_class(need_validate_class, class_):
    """
    Raises an exception if validate_class is not a
    subclass of adapter_class.

    :param str need_validate_class: The class to be validated.
    :param class class_: The class type to check against.


    :raise: InvalidTypeException
    """
    # If a dictionary was passed in, check if it has an import_path attribute
    if isinstance(need_validate_class, dict):
        if 'import_path' not in need_validate_class:
            raise InvalidTypeError(
                'The dictionary {} must contain a value for "import_path"'.format(
                    str(need_validate_class)
                )
            )

        # Set the class to the import path for the next check
        need_validate_class = need_validate_class.get('import_path')

    if not issubclass(import_module(need_validate_class), class_):
        raise InvalidTypeError(
            '{} must be a subclass of {}'.format(
                need_validate_class,
                class_.__name__
            )
        )


def get_function_arguments(fun):
    arguments = inspect.getfullargspec(fun)[0] or []
    defaults = inspect.getfullargspec(fun)[3] or ()
    args_len = len(arguments)
    defaults_len = len(defaults)

    # if it is a class method, then the self parameter is removed
    if 'self' in arguments:
        arguments.remove('self')

    all_argument_and_value = collections.OrderedDict()
    # all parameters without default values
    for argument in arguments[:(args_len-defaults_len)]:
        all_argument_and_value[argument] = None

    # all parameters with default values
    for argument, argument_default_value in zip(arguments[(args_len-defaults_len):], defaults):
        all_argument_and_value[argument] = argument_default_value

    return all_argument_and_value


def get_features(
        content,
        allow_pos=()):
    stop_word_file_path = os.path.join(PROJECT_DIR_PATH, 'data', 'stopwords.txt')
    jieba.analyse.set_stop_words(stop_word_file_path)
    # keyWord = jieba.analyse.textrank(content, topK=3, allowPOS=['ns', 'n', 'vn', 'v', 'nr'])
    return jieba.analyse.extract_tags(content, topK=max(3, int(len(content) / 6)), allowPOS=allow_pos)


def get_logger(name,
               level=logging.INFO,
               output_stream=sys.stdout,
               log_format='%(asctime)s [%(name)s] [%(filename)s:%(lineno)d] [%(levelname)s] [%(thread)d]: %(message)s'):
    """A function which writes formatted logging records to standard output or standard error output

    :rtype:
    :param str name: log name
    :param int level: Log level
    :param str output_stream: stream output
    :param str log_format: set the log format

    :return(class 'logging.Logger'): logging.Logger
    """
    # create logger
    logger_ = logging.getLogger(name)
    logger_.setLevel(level)

    # create handler
    screen_handler = logging.StreamHandler(stream=output_stream)
    file_handler = logging.FileHandler(
        os.path.join(PROJECT_DIR_PATH, 'log', LOG_FILE_NAME),
        mode='w',
        encoding='utf8'
    )
    # handler = logging.FileHandler(name, 'a', encoding='utf-8')
    screen_handler.setFormatter(logging.Formatter(log_format, datefmt='%Y-%m-%d %H:%M:%S'))
    file_handler.setFormatter(logging.Formatter(log_format, datefmt='%Y-%m-%d %H:%M:%S'))

    # add handler to logger
    logger_.addHandler(screen_handler)
    logger_.addHandler(file_handler)

    return logger_


logger = get_logger(APP_NAME, LOG_LEVEL)


def get_object_path(object_):
    module_name = inspect.getmodule(object_).__name__
    if inspect.isclass(object_) or inspect.isfunction(object_):
        return '{}.{}'.format(module_name, object_.__name__)
    elif inspect.ismethod(object_):
        return '{}.{}.{}'.format(module_name, object_.__self__.__class__.__name__, object_.__name__)
    else:
        path = module_name
        if hasattr(object_, '__class__'):
            path += '.' + object_.__class__.__name__

        if hasattr(object_, '__name__'):
            path += '.' + object_.__name__

        return path


def get_function_parameter_desc(fun):
    """Get the parameter description in the function"""
    all_parameter_desc = {}
    function_doc = inspect.getdoc(fun)
    if function_doc:
        for line in function_doc.split('\n'):
            line = line.strip()
            if line.startswith(':param'):
                parameter, desc = re.search(r':param\s+.*?(\S+):(.*)', line).groups()
                all_parameter_desc[parameter] = desc.strip()

    return all_parameter_desc


def echo(string):
    print(string)



from dateutil import parser as date_parser
from datetime import date
import forecastio
import time
from geocoder.api import get


baidumap_api_key = '5oEtquqXoPReYb4YLINasLikoCWRZLMr'
gaode_api_key = '38e6b376eac44085a8e72f6ef5e93fb0'
forecast_api_key = "3b1a0ead7333234fe0b428ab4058c4cf"


def get_weather(address, day):
    """
    :param str address: 所需要查询天气的城市
    :param str day: 需要查询天气的日期
    :return:
    """
    # 核查日期
    day = int(time.mktime(date_parser.parse(day).date().timetuple()))
    today = int(time.mktime(date.today().timetuple()))
    if day < today or day >= (today + 8 * 86400):
        return '只能查询8天内的天气情况'

    location = get(address, provider='gaode', method='geocode', key=gaode_api_key)

    # 通过经纬度获取天气信息
    forecast = forecastio.load_forecast(forecast_api_key, location.lat, location.lng)
    all_daily_weather = forecast.daily().data
    daily_weather = None
    for weather in all_daily_weather:
        if weather.d.get('time') == day:
            daily_weather = weather
            break

    if not daily_weather:
        return '无法查询到该日期的天气情况'

    return str({
        '详细地址': location.address,
        '最低温度': round(daily_weather.temperatureMin),
        '最高温度': round(daily_weather.temperatureMax),
        # '总结': daily_weather.summary,
        '降雨概率': '{}%'.format(int(daily_weather.precipProbability * 100)),
        '日期': time.strftime('%Y-%m-%d', time.localtime(daily_weather.d.get('time')))
    })


if __name__ == '__main__':
    for i in range(10000):
        print(get_weather('秦淮河', '20190726'))
        print(get_weather('夫子庙', '20190726'))
        print(get_weather('南京中山陵', '20190726'))
        print(get_weather('玄武湖', '20190726'))
        print(get_weather('台江区', '20190726'))
        print(get_weather('福州', '20190726'))
        print(get_weather('南京', '20190726'))





from .utils import logger, initialize_class, get_features, validate_class, import_module, get_object_path
from .extractor import Extractor
from .comparison import levenshtein_distance
from .exceptions import MethodNotImplementedError, ExtractDataError
from .constants import MAXIMUM_SIMILARITY_THRESHOLD, MINIMUM_SIMILARITY_THRESHOLD, NUMBER_OF_ANSWERS
from .storage import SQLStorage
from .conversation import Statement
from .models import statement_table_name, tag_table_name, tag_association_statement_table_name
from .tokenizer import jieba_segment as segment


class LogicAdapter:
    """
    This is an abstract class that represents the interface
    that all logic adapters should implement.

    :param storage:
        storage model
    :param maximum_similarity_threshold:
        The maximum amount of similarity between two statement that is required
        before the search process is halted. The search for a matching statement
        will continue until a statement with a greater than or equal similarity
        is found or the search set is exhausted.
        Defaults to 0.8
    :param default_response:
          The default response returned by this logic adapter
          if there is no other possible response to return.
    :type default_response: str or list or tuple
    """

    def __init__(self, **kwargs):
        storage = kwargs.get('storage', SQLStorage())

        # initialize storage class
        if isinstance(storage, str) or isinstance(storage, dict):
            self.storage = initialize_class(storage)
        else:
            self.storage = storage

        self.maximum_similarity_threshold = kwargs.get(
            'maximum_similarity_threshold', MAXIMUM_SIMILARITY_THRESHOLD
        )

        self.minimum_similarity_threshold = kwargs.get(
            'minimum_similarity_threshold', MINIMUM_SIMILARITY_THRESHOLD
        )

        self.number_of_answers = kwargs.get(
            'number_of_answers', NUMBER_OF_ANSWERS
        )

        self.default_responses = kwargs.get('default_response', [])

        # Convert a single string into a list
        if isinstance(self.default_responses, str):
            self.default_responses = [
                self.default_responses
            ]

        # logger
        self.logger = kwargs.get('logger', logger)

    def can_process(self, statement, **kwargs):
        """
        A preliminary check that is called to determine if a
        logic adapter can process a given statement. By default,
        this method returns true but it can be overridden in
        child classes as needed.

        :rtype: bool
        """
        return True

    def process(self, statement, **kwargs):
        """
        Override this method and implement your logic for selecting a response to an input statement.

        A confidence value and the selected response statement should be returned.
        The confidence value represents a rating of how accurate the logic adapter
        expects the selected response to be. Confidence scores are used to select
        the best response from multiple logic adapters.

        The confidence value should be a number between 0 and 1 where 0 is the
        lowest confidence level and 1 is the highest.

        :param str statement: An input statement to be processed by the logic adapter.

        :return Statement:
        """
        raise MethodNotImplementedError()

    def get_default_response(self, statement):
        """
        This method is called when a logic adapter is unable to generate any
        other meaningful response.
        """
        from random import choice

        if self.default_responses:
            response = Statement(
                question=statement,
                answer=choice(self.default_responses)
            )
        else:
            response = Statement(
                question=statement,
                answer=statement
            )

        # Set confidence to zero because a default response is selected
        response.confidence = 0

        return response

    @property
    def class_name(self):
        """
        Return the name of the current logic adapter class.
        This is typically used for logging and debugging.
        """
        return str(self.__class__.__name__)


class BestMatch(LogicAdapter):
    """
    A logic adapter that returns a response based on known responses to
    the closest matches to the input statement.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Use full library scan when there are no keywords
        self.using_full_library_scan = kwargs.get('using_full_library_scan', False)

        # comparator
        self.comparator = kwargs.get('comparator', levenshtein_distance)

    def can_process(self, statement, **kwargs):
        return not kwargs.get('context', {}).get('domain')

    def process(self, input_statement, **kwargs):
        features = get_features(input_statement)
        if features:
            self.logger.info('the keyword of the "{}" statement is "{}"'.format(
                input_statement, ','.join(['{}'.format(f) for f in features]))
            )
            sql = '''SELECT distinct s.* FROM {statement_table} s, {tag_table} t, {tag_association_statement_table} tas
                   where s.id = tas.statement_id
                   and t.id = tas.tag_id
                   and t.name in ({condition});'''.format(
                statement_table=statement_table_name,
                tag_table=tag_table_name,
                tag_association_statement_table=tag_association_statement_table_name,
                condition=','.join(['"{}"'.format(f) for f in features])
            )
            all_need_to_match_statements = self.storage.execute(sql)

            if not all_need_to_match_statements:
                if self.using_full_library_scan:
                    self.logger.info('unable to find database data using keywords, change to full library comparison')
                    all_need_to_match_statements = self.storage.all(statement_table_name)
                else:
                    self.logger.info('unable to find database data using keywords')
                    return []
            else:
                self.logger.info('use keywords to find "{}" pieces of data and use it for comparison'.format(
                    len(all_need_to_match_statements))
                )
        else:
            if self.using_full_library_scan:
                self.logger.info('"{}" statement has no keywords, full library comparison'.format(input_statement))
                all_need_to_match_statements = self.storage.all(statement_table_name)
            else:
                self.logger.info('"{}" statement has no keywords'.format(input_statement))
                return []

        # get similarity to all data in the database
        all_result = []
        for statement in all_need_to_match_statements:
            similarity = self.comparator(input_statement, statement.question)
            self.logger.debug(
                'the similarity between the statement "{}" and the statement "{}" is {}'.format(
                    input_statement, statement.question, similarity
                )
            )
            if similarity >= self.minimum_similarity_threshold:
                all_result.append(
                    Statement(
                        id=statement.id,
                        question=input_statement,
                        reference_question=statement.question,
                        confidence=similarity,
                        answer=statement.answer,
                        category=statement.category,
                        type=statement.type,
                        parameters=statement.parameters,
                        extractor=statement.extractor
                    )
                )

            # if the similarity is greater than the maximum_similarity_threshold,
            # the matching is stopped, which can reduce the matching time.
            if similarity >= self.maximum_similarity_threshold:
                return [all_result[-1]]

        # get the statement with the highest similarity
        res = []
        if all_result:
            all_result.sort(key=lambda s: s.confidence, reverse=True)
            res = all_result[:self.number_of_answers]
        # else:
        #     res = [self.get_default_response(input_statement)]

        return res


class DomainManager(LogicAdapter):
    def can_process(self, statement, **kwargs):
        return kwargs.get('context', {}).get('domain')

    def process(self, input_statement, **kwargs):
        context = kwargs.get('context', {})
        id_ = context.get('id')
        question = context.get('question')
        reference_question = context.get('reference_question')
        confidence = context.get('confidence')
        answer = context.get('answer')
        category = context.get('category')
        type_ = context.get('type')
        parameters = context.get('parameters')
        need_extract_parameter = context.get('need_extract_parameter')
        extractor = context.get('extractor')

        # initial extractor
        validate_class(extractor, Extractor)
        extractor_obj = import_module(extractor)(input_statement)

        # extract parameter
        get_arg_val_fun = getattr(extractor_obj, 'extract_{}'.format(need_extract_parameter),
                                  None) or getattr(extractor_obj, 'extract_general')
        self.logger.info(
            'get the value of the "{}" parameter using "{}" method'.format(
                need_extract_parameter, get_object_path(get_arg_val_fun)
            )
        )
        try:
            argument_value = get_arg_val_fun()
            parameters[need_extract_parameter] = argument_value
            self.logger.info('the value is "{}"'.format(argument_value))
        except ExtractDataError:
            self.logger.error(
                'The "{}" method can not get the value of the "{}" parameter from the context "{}".'.format(
                    get_object_path(get_arg_val_fun),
                    need_extract_parameter,
                    input_statement
                )
            )

        return [
            Statement(
                id=id_,
                question=question,
                reference_question=reference_question,
                confidence=confidence,
                answer=answer,
                category=category,
                type=type_,
                parameters=parameters,
                extractor=extractor
            )
        ]


class WhatCanIDo(LogicAdapter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # comparator
        self.comparator = kwargs.get('comparator', levenshtein_distance)

    def can_process(self, statement, **kwargs):
        if not kwargs.get('context', {}).get('domain'):
            segment_text = segment.cut(statement, seg_only=True, remove_stop_word=True)
            for keyword in ['功能', '能力']:
                if keyword in segment_text:
                    return True

    def process(self, input_statement, **kwargs):
        questions = ['机器人功能', '机器人能力']
        for question in questions:
            similarity = self.comparator(input_statement, question)
            if similarity >= self.maximum_similarity_threshold:
                answer = '我提供的服务如下:\n'
                index = 0
                for dynamic_qa in self.storage.filter(statement_table_name, type=1):
                    index += 1
                    answer += '{}. {}\n'.format(index, dynamic_qa.question)

                # answer += '{}. 我的大脑还植入了{}条与门户有关的知识,你可以问我关于这方面的问题\n'.format(
                #     index + 1,
                #     self.storage.count(statement_table_name) - index
                # )
                return [
                    Statement(
                        id=-1,
                        question=input_statement,
                        reference_question=question,
                        confidence=similarity,
                        answer=answer,
                        category='',
                        type=0,
                        parameters='',
                        extractor=''
                    )
                ]

        return []



from sqlalchemy import Column, Integer, String, DateTime, ForeignKey  # Text, Table
# from sqlalchemy.orm import relationship
from sqlalchemy.sql import func
from sqlalchemy.ext.declarative import declared_attr, declarative_base


class ModelBase:
    """
    An augmented base class for SqlAlchemy models.
    """

    @declared_attr
    def __tablename__(cls):
        """
        Return the lowercase class name as the name of the table.
        """
        table_name = ''
        for index, char in enumerate(cls.__name__):
            if index == 0:
                table_name += char.lower()
            else:
                if char.isupper():
                    table_name += '_{}'.format(char.lower())
                else:
                    table_name += char

        return table_name

    id = Column(
        Integer,
        primary_key=True,
        autoincrement=True
    )


Base = declarative_base(cls=ModelBase)


class Statement(Base):
    question = Column(
        String(length=1000),
        unique=True
    )
    answer = Column(
        String(length=1000),
        nullable=False
    )
    category = Column(
        String(length=100),
        server_default='其他'
    )
    type = Column(
        Integer,
        server_default='0'
    )
    parameters = Column(
        String(length=255),
        server_default=''
    )
    extractor = Column(
        String(length=255),
        server_default=''
    )
    create_at = Column(
        DateTime(timezone=True),
        server_default=func.now()
    )


class Tag(Base):
    """
    A tag that describes a statement.
    """

    name = Column(
        String(length=255),
        unique=True
    )


class TagAssociationStatement(Base):
    tag_id = Column(
        Integer,
        ForeignKey('tag.id')
    )
    statement_id = Column(
        Integer,
        ForeignKey('statement.id')
    )


class AccessLog(Base):
    statement_id = Column(
        Integer,
        # ForeignKey('statement.id')
    )
    access_time = Column(
        DateTime(timezone=True),
        server_default=func.now()
    )


statement_table_name = Statement.__tablename__
tag_table_name = Tag.__tablename__
tag_association_statement_table_name = TagAssociationStatement.__tablename__
access_log_table_name = AccessLog.__tablename__



from geopy.geocoders import Nominatim
from geopy.exc import GeocoderTimedOut, GeocoderUnavailable
from dateutil import parser as date_parser
from datetime import date
import forecastio
import re
import time


api_key = "3b1a0ead7333234fe0b428ab4058c4cf"


class AddressError(Exception):
    """
    An exception to be raised when the address incorrect
    """

    def __init__(self, address):
        self.message = '地址"{}"输入有误'.format(address)
        super().__init__(self.message)


def get_weather(address, day):
    """
    :param str address: 地址
    :param str day: 时间, 格式为YYYYMMDD
    :return:
    """
    # 核查日期
    day = int(time.mktime(date_parser.parse(day).date().timetuple()))
    today = int(time.mktime(date.today().timetuple()))
    if day < today or day >= (today + 8 * 86400):
        return '只能查询8天内的天气情况'

    gps = Nominatim()
    location = None
    detail_address = ''
    # 获取地址的经纬度
    for i in range(5):
        try:
            location = gps.geocode(address)
        except (GeocoderTimedOut, GeocoderUnavailable):
            continue

        break

    if not location:
        raise AddressError(address)

    # 拼接详细地址信息
    for i in reversed(location.address.split(',')):
        match = re.search(r'[\u4e00-\u9fff]+', i)
        if match:
            detail_address += match.group()

    # 通过经纬度获取天气信息
    forecast = forecastio.load_forecast(api_key, location.latitude, location.longitude)
    all_daily_weather = forecast.daily().data
    daily_weather = None
    for weather in all_daily_weather:
        if weather.d.get('time') == day:
            daily_weather = weather
            break

    if not daily_weather:
        return '无法查询到该日期的天气情况'

    return {
        '详细地址': detail_address,
        '最低温度': round(daily_weather.temperatureMin),
        '最高温度': round(daily_weather.temperatureMax),
        # '总结': daily_weather.summary,
        '降雨概率': '{}%'.format(int(daily_weather.precipProbability * 100)),
        '日期': time.strftime('%Y-%m-%d', time.localtime(daily_weather.d.get('time')))
    }


if __name__ == '__main__':
    print(get_weather('青城山', '20190725'))
    print(get_weather('玄武湖', '20190725'))
    print(get_weather('秦淮河', '20190725'))
    print(get_weather('南京市雨花台区', '20190725'))
    print(get_weather('福州', '20190725'))



class ChatbotError(Exception):
    """
    chatbot base error
    """


class InvalidTypeError(ChatbotError):
    """
    An exception to be raised when an adapter
    of an unexpected class type is received.
    """
    pass


class MethodNotImplementedError(NotImplementedError, ChatbotError):
    """
    An exception to be raised when an extractor method has not been implemented.
    Typically this indicates that the developer is expected to implement the
    method in a subclass.
    """

    def __init__(self, message='This method must be overridden in a subclass method.'):
        """
        Set the message for the exception.
        """
        self.message = message
        super().__init__(message)


class ExtractDataError(ChatbotError):
    """
    An exception to be raised when the extractor cannot extract the parameter value
    """

    def __init__(self, parameter_name):
        """
        Set the message for the exception.
        """
        # current_method_name = inspect.stack()[1][3]
        self.parameter = parameter_name
        self.message = 'unable to extract {} data'.format(parameter_name)
        super().__init__(self.message)


class NotEnoughParameterError(ChatbotError):
    def __init__(self, parameter_name, message=None):
        self.parameter = parameter_name
        if message:
            self.message = '请提供{}'.format(message)
        else:
            self.message = '请提供{}参数的值'.format(parameter_name)
        super().__init__(self.message)


class NotFoundExtractorMethod(ChatbotError):
    def __init__(self, method):
        self.message = 'The "{}" method does not exist.'.format(method)
        super().__init__(self.message)


class StorageError(ChatbotError):
    """"""


class ModelNotExistError(StorageError):
    """
    An exception to be raised when the model does not exist in the database
    """

    def __init__(self, model_name):
        """
        Set the message for the exception.
        """
        self.message = 'the "{}" model does not exist in the database'.format(model_name)
        super().__init__(self.message)


class DeleteDataWithoutConditionError(StorageError):
    """
    An exception to be raised when deleting database data does not provide conditions
    """

    def __init__(self, message='deleting database data does not provide conditions'):
        """
        Set the message for the exception.
        """
        self.message = message
        super().__init__(message)


class UniqueError(StorageError):
    def __init__(self, message='line already exists, violates uniqueness constraint'):
        """
        Set the message for the exception.
        """
        self.message = message
        super().__init__(message)


class ExecuteSqlError(StorageError):
    def __init__(self, message='line already exists, violates uniqueness constraint'):
        """
        Set the message for the exception.
        """
        self.message = message
        super().__init__(message)



"""
chatbot constants
"""
import logging
import os

"""
application veriosn
"""
APP_VERSION = '2.7'

"""
application name
"""
APP_NAME = 'chatbot'

"""
log level
"""
LOG_LEVEL = logging.INFO

"""
log file name
"""
LOG_FILE_NAME = 'chatbot.log'

"""
The maximum amount of similarity between two statement that is required
before the search process is halted. The search for a matching statement
will continue until a statement with a greater than or equal similarity
is found or the search set is exhausted.
"""
MAXIMUM_SIMILARITY_THRESHOLD = 0.85

"""
The minimum amount of similarity between two statement, below this value, 
the robot cannot answer questions from the user.
"""
MINIMUM_SIMILARITY_THRESHOLD = 0.2

"""
the number of answers each adapter has to answer
"""
NUMBER_OF_ANSWERS = 5


"""
project path,build paths inside the project like this: os.path.join(BASE_DIR, ...)
"""
PROJECT_DIR_PATH = os.path.dirname((os.path.abspath(__file__)))

"""
database address
"""
DEFAULT_DATABASE_URI = 'sqlite:///db.sqlite3'

"""
Get the maximum number of errors allowed by the context parameter. 
If the number of times exceeds, the problem is terminated.
"""
CONTEXT_PARAMETER_MAX_ERROR_COUNT = 3


import sys
sys.path.append(r'E:\pycharm_projects')
from cloudoneapi import CloudOne


def get_all_project():
    cloudone_client = CloudOne('http://10.1.196.76:8080/', 'chenxb', '88888888')
    try:
        return cloudone_client.get_all_project_info()
    except Exception as e:
        return e.args


def get_all_job(project_name):
    cloudone_client = CloudOne('http://10.1.196.76:8080/', 'chenxb', '88888888')
    try:
        return cloudone_client.get_all_job(project_name)
    except Exception as e:
        return e.args

def get_all_job_by_type(project_name, job_type):
    cloudone_client = CloudOne('http://10.1.196.76:8080/', 'chenxb', '88888888')
    try:
        return cloudone_client.get_all_job_by_type(project_name, job_type)
    except Exception as e:
        return e.args


if __name__ == '__main__':
    print(get_all_job('test111'))
    print(get_all_project())
    print(get_all_job_by_type('test111', 'SHELL脚本执行'))


from .tokenizer import jieba_segment as segment
from difflib import SequenceMatcher


class Comparator:
    def __call__(self, statement_a, statement_b):
        return self.compare(statement_a, statement_b)

    @staticmethod
    def _cut_statement(statement):
        return segment.cut(statement, seg_only=True)

    def compare(self, statement_a, statement_b):
        return 0


class LevenshteinDistance(Comparator):
    """
    Compare two statements based on the Levenshtein distance
    of each statement.
    """

    def compare(self, statement, other_statement):
        """
        Compare the two input statements.

        :return float: The percent of similarity between the text of the statements.
        """

        # Return 0 if either statement has a falsy text value
        if not statement or not other_statement:
            return 0

        similarity = SequenceMatcher(
            None,
            statement,
            other_statement
        )

        # Calculate a decimal percent of the similarity
        percent = round(similarity.ratio(), 2)

        return percent


# class BM25Distance(Comparator):
#     """
#     Compare the similarity of two sentences using the bm25 algorithm
#
#     :return float: The percent of similarity between the text of the statements.
#     """
#
#     def compare(self, statement, other_statement):
#         # Return 0 if either statement has a falsy text value
#         if not statement or not other_statement:
#             return 0
#
#         # Calculate a decimal percent of the similarity
#         statement_token = self._cut_statement(statement)
#         other_statement_token = self._cut_statement(other_statement)
#         s = SnowNLP(statement_token)
#         percent = s.sim(other_statement_token)
#
#         return percent


# class VectorDistance(Comparator):
#     """
#     Calculate the similarity of two statements.
#     """
#     def __init__(self, **kwargs):
#         # logger
#         self.logger = kwargs.get('logger', logger)
#         app_id = '16082031'
#         api_key = 'iNCC8w5mI63I7eYh0Y1t4Spv'
#         secret_key = 'rjOM3cERogv4RqDEux0n8nbLBItOP7iF'
#
#         self._baidu_client = AipNlp(app_id, api_key, secret_key)
#
#     def _get_word_vector(self, word):
#         response = self._baidu_client.wordEmbedding(word)
#         if not response.get('vec'):
#             self.logger.debug('dictionary does not have this word "{}"'.format(word))
#         return response.get('vec', [0])
#
#     def _get_statement_vector(self, statement):
#         statement_vector = []
#         print(self._cut_statement(statement))
#         for word in self._cut_statement(statement):
#             word_vector = self._get_word_vector(word)
#             if word_vector != [0]:
#                 statement_vector.append(word_vector)
#
#         return statement_vector if statement_vector else [[0]]
#
#     def compare(self, statement, other_statement):
#         """
#         Compare the two input statements.
#
#         :return: The percent of similarity between the vector distance.
#         :rtype: float
#         """
#         # Return 0 if either statement has a falsy text value
#         if not statement or not other_statement:
#             return 0
#
#         statement_vector = self._get_statement_vector(statement)
#         other_statement_vector = self._get_statement_vector(other_statement)
#         return dot(matutils.unitvec(array(statement_vector).mean(axis=0)),
#                    matutils.unitvec(array(other_statement_vector).mean(axis=0)))


levenshtein_distance = LevenshteinDistance()
# vector_distance = VectorDistance()



"""
Statement pre-processors.
"""


def clean_whitespace(statement):
    """
    Remove any consecutive whitespace characters from the statement
    """
    import re

    # Replace linebreaks and tabs with spaces
    statement = statement.replace('\n', ' ').replace('\r', ' ').replace('\t', ' ')

    # Remove any leeding or trailing whitespace
    statement = statement.strip()

    # Remove consecutive spaces
    statement = re.sub(' +', ' ', statement)

    return statement


def unescape_html(statement):
    """
    Convert escaped html characters into unescaped html characters.
    For example: "&lt;b&gt;" becomes "<b>".
    """
    import html

    statement = html.unescape(statement)

    return statement



import re
# import inspect
import jieba
from .tokenizer import jieba_segment as segment
from .exceptions import ExtractDataError
from .utils import logger


class Extractor:
    def __init__(self, text, **kwargs):
        self._text = text
        # logger
        self.logger = kwargs.get('logger', logger)

    def __str__(self):
        return self.__class__.__name__

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        self._text = value

    @staticmethod
    def _add_word_to_jieba(word_list):
        for word in word_list:
            jieba.add_word(word)

    @staticmethod
    def _del_word_from_jieba(word_list):
        for word in word_list:
            jieba.del_word(word)

    def extract(self,
                key,
                expected_value_pattern=r'[^\u4e00-\u9fff]+',
                tags=('eng', 'n', 'nr', 'ns', 'nt', 'nw', 'nz', 'vn', 'x')):
        """extract data

        :param tuple key: keyword
        :param str expected_value_pattern: expect pattern， The default is non-Chinese
        :param tuple tags: part of speech
        :return str:

        Example:

        ex = Extractor('获取ip地址为1.1.1.1的主机的磁盘空间使用率')
        ip1 = ex.extract(key=('ip地址', 'ip'), expected_value_pattern=r'([0-9]{1,3}.){3}[0-9]{1,3}')
        ex.text = '获取1.1.1.1主机的磁盘空间使用率'
        ip2 = ex.extract(key=('主机', 'ip地址', 'ip'), expected_value_pattern=r'([0-9]{1,3}.){3}[0-9]{1,3}')
        print(ip1, ip2)
        """
        # add word to jieba
        self._add_word_to_jieba(key)

        # segment
        segment_text = segment.cut(self.text, seg_only=False, remove_stop_word=True)
        self.logger.debug('"{}" statement after the word segmentation is "{}"'.format(self.text, segment_text))
        key_index = -1
        for index, word_and_tag in enumerate(segment_text):
            if word_and_tag[0] in key:
                key_index = index
                break

        if key_index != -1:
            the_word_index_before_keyword = key_index - 1
            the_word_index_after_keyword = key_index + 1
            if the_word_index_before_keyword >= 0 and \
                    re.match(expected_value_pattern + r'$', segment_text[the_word_index_before_keyword][0]) and \
                    segment_text[the_word_index_before_keyword][1] in tags:
                return segment_text[the_word_index_before_keyword][0]

            if the_word_index_after_keyword < len(segment_text) and \
                    re.match(expected_value_pattern + r'$', segment_text[the_word_index_after_keyword][0]) and \
                    segment_text[the_word_index_after_keyword][1] in tags:
                return segment_text[the_word_index_after_keyword][0]

        raise ExtractDataError(key[0])

    def extract_general(self):
        return self.text.strip()


class CloudOneExtractor(Extractor):
    def extract_project_en_name(self):
        return self.extract(
            key=('项目', '项目名', '项目名称', '项目英文名', '项目英文名称', 'project', 'project_name', 'project_en_name')
        )

    def extract_job_type(self):
        return self.extract(
            key=('任务类别', '类别', 'job_type'),
            expected_value_pattern='.*',
            tags=('x',)
        )

    extract_project_name = extract_project_en_name


class HostExtractor(Extractor):
    def extract_ip(self):
        try:
            ip = self.extract(
                key=('ip地址', 'ip'), expected_value_pattern=r'(\d{1,3}\.){3}\d{1,3}', tags=('m',)
            )
        except ExtractDataError as e:
            ip_match = re.search(r'(\d{1,3}\.){3}\d{1,3}', self.text)
            if ip_match:
                ip = ip_match.group()
            else:
                raise e
        return ip

    def extract_username(self):
        try:
            username = self.extract(
                key=('用户名', '用户', 'username', 'user'), expected_value_pattern=r'[^\u4e00-\u9fff]+'
            )
        except ExtractDataError as e:
            username_match = re.search(r'[^\u4e00-\u9fff]+', self.text)
            if username_match:
                username = username_match.group()
            else:
                raise e
        return username

    def extract_password(self):
        try:
            password = self.extract(
                key=('密码', 'password'), expected_value_pattern=r'[^\u4e00-\u9fff]+'
            )
        except ExtractDataError as e:
            password_match = re.search(r'[^\u4e00-\u9fff]+', self.text)
            if password_match:
                password = password_match.group()
            else:
                raise e
        return password


class WeatherExtractor(Extractor):
    def extract_address(self):
        try:
            address = self.extract(
                key=('地址', 'address'), expected_value_pattern=r'[\u4e00-\u9fff]+'
            )
        except ExtractDataError as e:
            address_match = re.search(r'[\u4e00-\u9fff]+', self.text)
            if address_match:
                address = address_match.group()
            else:
                raise e
        return address

    def extract_day(self):
        try:
            day = self.extract(
                key=('时间', '日期', 'day'), expected_value_pattern=r'\d{8}'
            )
        except ExtractDataError as e:
            day_match = re.search(r'\d{8}', self.text)
            if day_match:
                day = day_match.group()
            else:
                raise e
        return day



from setuptools import setup
import os


PROJECT_ROOT, _ = os.path.split(__file__)
PROJECT_NAME = 'chatbot-py'
VERSION = '2.72'
PROJECT_AUTHORS = 'lb'
PROJECT_EMAILS = '423497786@qq.com'
PROJECT_URL = 'https://github.com/lin423497786/chatbot-py'
SHORT_DESCRIPTION = (
    '中文聊天机器人， 根据自己的语料训练出自己想要的聊天机器人，可以用于智能客服、在线问答、智能聊天等场景!'
)


# Get the long description from the relevant file
try:
    DESCRIPTION = open(os.path.join(PROJECT_ROOT, 'README.rst')).read()
except IOError:
    DESCRIPTION = SHORT_DESCRIPTION


REQUIREMENTS = []
with open('requirements.txt') as requirements:
    for requirement in requirements:
            REQUIREMENTS.append(requirement)


setup(
    name=PROJECT_NAME.lower(),
    version=VERSION,
    author=PROJECT_AUTHORS,
    author_email=PROJECT_EMAILS,
    url=PROJECT_URL,
    description=SHORT_DESCRIPTION,
    long_description=DESCRIPTION,
    platforms=['any'],
    keywords=['chatbot'],
    python_requires='>=3.4, <4',
    install_requires=REQUIREMENTS,

    # packages=find_packages(exclude=['contrib', 'docs', 'test*']),
    packages=[
        'chatbot',
        'chatbot.ext',
    ],
    package_dir={'chatbot': 'chatbot'},
    test_suite='tests',
    package_data={
        'chatbot': [
            '*.*',
            'log/*',
            'data/*',
        ],
    }
)



from .adapter import LogicAdapter
from .utils import initialize_class, logger, validate_class, import_module, get_features, get_object_path
from .constants import MAXIMUM_SIMILARITY_THRESHOLD, MINIMUM_SIMILARITY_THRESHOLD, \
    NUMBER_OF_ANSWERS, CONTEXT_PARAMETER_MAX_ERROR_COUNT
from .exceptions import NotEnoughParameterError
from .models import statement_table_name, tag_table_name, tag_association_statement_table_name, access_log_table_name
from .storage import SQLStorage


class ChatBot:
    def __init__(self, name, **kwargs):
        self.name = name
        # storage = kwargs.get('storage', 'chatbot.storage.SQLStorage')
        storage = kwargs.get('storage', SQLStorage())

        # initialize storage class
        if isinstance(storage, str) or isinstance(storage, dict):
            self.storage = initialize_class(storage)
        else:
            self.storage = storage

        # logger
        self.logger = kwargs.get('logger', logger)

        # Stop searching when it is greater than the match
        self.maximum_similarity_threshold = kwargs.get('maximum_similarity_threshold', MAXIMUM_SIMILARITY_THRESHOLD)

        # The minimum amount of similarity between two statement, below this value,
        # the chatbot cannot answer questions from the user.
        self.minimum_similarity_threshold = kwargs.get(
            'minimum_similarity_threshold', MINIMUM_SIMILARITY_THRESHOLD
        )

        # number of answers
        self.number_of_answers = kwargs.get(
            'number_of_answers', NUMBER_OF_ANSWERS
        )

        # Logic adapters used by the chatbot
        adapters = kwargs.get('logic_adapters', [
            {
                'import_path': 'chatbot.adapter.WhatCanIDo',
                'storage': self.storage,
                'logger': self.logger
            },
            {
                'import_path': 'chatbot.adapter.DomainManager',
                'logger': self.logger
            },
            {
                'import_path': 'chatbot.adapter.BestMatch',
                'storage': self.storage,
                'logger': self.logger
            },
        ])

        # initialize the adapter class
        self.logic_adapters = []
        for adapter in adapters:
            validate_class(adapter, LogicAdapter)
            logic_adapter = initialize_class(adapter)
            self.logic_adapters.append(logic_adapter)

        # the processing before the statement is passed to the chatbot
        self.preprocessors = []
        preprocessors = kwargs.get(
            'preprocessors', [
                'chatbot.preprocessor.clean_whitespace'
            ]
        )
        for preprocessor in preprocessors:
            self.preprocessors.append(import_module(preprocessor))

        if kwargs.get('initialize', True):
            self.initialize()

    def initialize(self):
        pass

    def get_response(self, input_statement=None, **kwargs):
        """
        Return the bot's response based on the input.

        :param str input_statement: string
        :returns str: a response to the input
        """
        self.logger.info('the processing the statement "{}"'.format(input_statement))
        # preprocess the input statement
        for preprocessor in self.preprocessors:
            input_statement = preprocessor(input_statement)
            self.logger.info('after the preprocessor "{}" processing, the statement becomes "{}"'.format(
                get_object_path(preprocessor), input_statement
            ))

        response = {
            'text': '',
            'context': kwargs.get('context', {'domain': False}),
        }
        # matching statements for each adapter
        all_adapter_answers = []
        for adapter in self.logic_adapters:
            if adapter.can_process(input_statement, **kwargs):
                self.logger.info('"{}" adapter starts matching'.format(get_object_path(adapter)))
                answers = adapter.process(input_statement, **kwargs)
                all_adapter_answers.extend(answers)
                self.logger.info('"{}" adapter select "{}" answers'.format(get_object_path(adapter), len(answers)))
                for answer in answers:
                    self.logger.info(
                        '"{}" adapter select the answer to the "{}" question as a reply, the confidence is {}'.format(
                            get_object_path(adapter), answer.reference_question, answer.confidence
                        )
                    )
                    # stop matching
                    if answer.confidence >= self.maximum_similarity_threshold:
                        self.logger.info(
                            'the similarity is {} higher than the {} parameter, stop matching.'.format(
                                answer.confidence, 'maximum_similarity_threshold'
                            )
                        )
                        try:
                            response['text'] = answer.get_answer()
                            response['context'] = {'domain': False}
                            # record access log
                            if answer.id != -1:
                                self.storage.create(model_name=access_log_table_name, statement_id=answer.id)
                        except NotEnoughParameterError as e:
                            if response['context'].get('need_extract_parameter') == e.parameter:
                                error_count = response['context'].get('need_extract_parameter_count', 0) + 1
                            else:
                                error_count = 0

                            if error_count >= CONTEXT_PARAMETER_MAX_ERROR_COUNT:
                                response['text'] = '错误次数过多,该问题已被终止!'
                                response['context']['domain'] = False
                            else:
                                response['text'] = e.message
                                response['context'] = answer.serialize()
                                response['context']['need_extract_parameter'] = e.parameter
                                response['context']['need_extract_parameter_count'] = error_count
                                response['context']['domain'] = True
                        self.logger.info(
                            'finally the response of the "{}" statement is "{}"'.format(input_statement, response)
                        )
                        return response
            else:
                self.logger.info(
                    'not processing the statement using "{}"'.format(get_object_path(adapter))
                )

        all_adapter_answers.sort(key=lambda s: s.confidence, reverse=True)
        all_adapter_answers = all_adapter_answers[:self.number_of_answers]
        if all_adapter_answers:
            response['text'] = '看看这些内容对您有帮助么？\n{}\n都不是？请用一句话完整描述您的问题'.format(
                '\n'.join(['{}.{}'.format(index + 1, answer.reference_question)
                           for index, answer in enumerate(all_adapter_answers)])
            )
        else:
            response['text'] = '很抱歉，没有理解您的意思，请用简短的话描述您的问题，比如"获取主机的磁盘空间使用率？"'

        self.logger.info('finally the response of the "{}" statement is "{}"'.format(
            input_statement, response
        ))

        return response

    def learn(self, question, answer, category='其他', type_=0, parameters=None, extractor=None):
        """
        Learn that the statement provided is a valid response.
        """
        features = get_features(question)
        if not features:
            message = 'statement "{}" has no features, so skip learning'.format(question)
            self.logger.warning(message)
            return False, message

        # add data to statement table
        try:
            statement_id = self.storage.create(
                model_name=statement_table_name,
                question=question,
                answer=answer,
                category=category,
                type=type_,
                parameters=parameters,
                extractor=extractor
            )

            # add data to tag table and tag_association_statement table
            for feature in features:
                tag = list(self.storage.filter(model_name=tag_table_name, name=feature))
                if len(tag) == 0:
                    tag_id = self.storage.create(
                        model_name=tag_table_name,
                        name=feature
                    )
                else:
                    tag_id = tag[0].id

                self.storage.create(
                    model_name=tag_association_statement_table_name,
                    tag_id=tag_id,
                    statement_id=statement_id
                )
        except Exception as e:
            message = 'inserting data into the database failed, {}'.format(e)
            self.logger.warning(message)
            return False, message

        self.logger.info('add "{}" as the answer to "{}"'.format(
            answer,
            question
        ))
        return True, ''

    def forget(self, question):
        """
        Delete data related to the question from the database
        """
        try:
            all_statement = self.storage.filter(
                model_name=statement_table_name,
                question=question
            )

            # Delete the data of tag_association_statement_table_name table
            for statement in all_statement:
                self.storage.delete(
                    model_name=tag_association_statement_table_name,
                    statement_id=statement.id
                )

            # Delete the data of statement_table_name table
            self.storage.delete(model_name=statement_table_name, question=question)
        except Exception as e:
            message = 'failed to delete data from the database, {}'.format(e)
            self.logger.warning(message)
            return False, message

        self.logger.info('delete "{}" question data from the database table'.format(question))
        return True, ''



import os
import codecs
import jieba.posseg as pseg
import jieba
from .utils import logger
from .constants import PROJECT_DIR_PATH


class Tokenizer:
    def __init__(self, **kwargs):
        stop_word_file_path = kwargs.get('stop_word_file_path',
                                         os.path.join(PROJECT_DIR_PATH, 'data', 'stopwords.txt'))
        self._all_stop_word = list(self.get_stop_word(stop_word_file_path))

        # logger
        self.logger = kwargs.get('logger', logger)

    @staticmethod
    def get_stop_word(file_path):
        with codecs.open(file_path, 'r', encoding='utf8') as f:
            for line in f:
                yield line.strip()


class JiebaTokenizer(Tokenizer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # load custom dictionary
        custom_dictionary_file_path = kwargs.get('custom_dictionary_file_path',
                                                 os.path.join(PROJECT_DIR_PATH, 'data', 'dict.txt'))
        jieba.load_userdict(custom_dictionary_file_path)

    def cut(self, statement, seg_only=False, remove_stop_word=False):
        segment_statement = []
        if seg_only:
            for word in jieba.cut(statement):
                if remove_stop_word:
                    if word not in self._all_stop_word:
                        segment_statement.append(word)
                else:
                    segment_statement.append(word)
        else:
            for word, tag in pseg.cut(statement):
                if remove_stop_word:
                    if word not in self._all_stop_word:
                        segment_statement.append((word, tag))
                else:
                    segment_statement.append((word, tag))

        return segment_statement


jieba_segment = JiebaTokenizer()
